structure Sasl :> SASL =
struct
  structure CCalls = CCalls(structure CCInfo = CCInfoDefaults)
  structure C_Sasl = C_SASL(structure C = CCalls)

  open C_Sasl

  type conn = connection

  structure R = Response

  (* val init : unit -> unit *)
  fun init () = sasl_server_init "legcap"

  val listmechs = sasl_listmech ("(SASL \"", "\" \"", "\")")

  (* val newconn : unit -> conn *)
  fun newconn () =
    (let
      val secflags = sasl_build_secflags ()
      val conn = sasl_server_new "acap" (secflags)
     in
       sasl_set_security_properties conn (0, 0, 0, secflags);
       conn
     end handle exn => (print ("exception: " ^ (exnMessage exn) ^ "\n");
			raise exn))

  exception ExpectedInput

  (* returns NONE on failure, SOME name on success *)
  (* val doauth : conn -> (string * string * string option) -> 
                       -> ((unit -> string option) 
                                  * Response.response Mailbox.mbox)
                       -> string option *)
  fun doauth conn (tag, mech, first) (response, outch) =
    (let
       fun format s = "{" ^ (Int.toString (size s)) ^ "}\r\n" ^ s
	 
       fun f (OK s) = Mailbox.send (outch, R.OK(SOME tag, [R.SASL (format s)], 
						"welcome"))
	 | f (CONTINUE s) = 
	 let 
	   val _ = print ("sending '" ^ s ^ "'\n")
	   val _ = Mailbox.send (outch, R.CONTINUE(format s))
	     
	   val s = case response () of
	     (SOME s) => (print ("got '" ^ s ^ "'\n"); s)
	   | NONE => raise ExpectedInput
	 in
	   f (sasl_server_step conn s)
	 end
     in
       f (sasl_server_start conn mech first);
       SOME (sasl_get_username conn) 
     end) handle ExpectedInput => (fail (outch, tag); NONE)
               | (exn as SaslError s) => (print ("got saslerror " ^ s ^ "\n");
					  fail (outch, tag); NONE)
    
  and fail (outch, tag) = Mailbox.send (outch, R.NO(SOME tag, nil,
						    "not authenticated"))

  (* val doneconn : conn -> unit *)
  fun doneconn conn = sasl_dispose conn

  (* val done : unit -> unit *)
  fun done () = sasl_done ()

end
