structure Value :> VALUE = 
struct
  datatype value =
    Single of string
  | List of string list
  | Nil
  | Default

  exception DefaultValue

  (* this would also be legal to just look for " \ CR LF *)
  fun safeChar c = (Char.isPrint c) andalso c <> #"\"" andalso c <> #"\\"

  fun stringAll f = 
    let
      fun g NONE = true
	| g (SOME (c, ss)) = f c andalso g (Substring.getc ss)
    in
      g o Substring.getc o Substring.all
    end

  val safe = stringAll safeChar

  fun wrap s = if (size s < 1000) andalso safe s then ("\"" ^ s ^ "\"")
	       else ("{" ^ (Int.toString (size s)) ^ "}\r\n" ^ s)

  fun toString (Single s) = wrap s
    | toString (List nil) = toString (Nil)
    | toString (List (a::l)) = "(" ^ (wrap a) ^
             (List.foldl (fn (s, acc) => (" " ^ (wrap s) ^ acc)) ")" l)
    | toString (Nil) = "NIL"
    | toString (Default) = raise DefaultValue

  fun size (Single s) = [String.size s]
    | size (List l) = List.map String.size l
    | size Nil = nil
    | size Default = nil

end
